/*
 * tmscore.h
 *
 *  Created on: Feb 7, 2012
 *      Author: jianjun
 */

#ifndef TMSCORE_H_
#define TMSCORE_H_

extern"C" {
float gdtts_(char modelname [], char nativename []);
}

/* Store a protein structures */
class TmObj
{
	/* The file name of a PDB file */
	string fileName;
public:

	TmObj() {
	}

	TmObj(const TmObj & obj) {
		copyConstruct(obj);
	}

	~TmObj() {
	}

	void copyConstruct(const TmObj & obj) {
		fileName = obj.fileName;
	}

	TmObj & operator=(const TmObj & obj) {
		copyConstruct(obj);
		return *this;
	}

	int operator==(const TmObj & obj) {
		if(fileName == obj.fileName) return 1;
		else return 0;
	}

	string getFileName() const {
		return fileName;
	}

	void print(ostream & os) const {
		os << "fileName " << fileName << endl;
	}

	/* Read in a data string */
	const void readin(string & data) {
		fileName = data;
	}

	/* (Do nothing) Rarse a data file to read in the structure */
	void loadData()
	{
	}

	/* Parse a line for tmscore result and return the GDT-TS value.
	 * Return -1 if doesn't contain a GDT-TS score*/
	double getGdtTs(std::string line) {
		double retVal = 1;

		if(line.size() > 0) {
			std::vector<std::string> temp;
			//boost::split(temp, line, boost::is_any_of(" "), boost::token_compress_on);
			Pattern::split(temp, line, " ");
			if (temp.size() > 0 && temp[0] == "GDT-TS-score=") {
				string trimmed = Pattern::trim(temp[1]);
				retVal = - atof(trimmed.c_str());
			}
		}
		return retVal;
	}

	/* Parse a line for tmscore result and return the GDT-TS value.
	 * Return -1 if doesn't contain a GDT-TS score*/
	double getGdtTs(char line[]) {
		double retVal = 1;

		if(sizeof line > 0) {
			char * pch;
			pch = strtok(line, " ");
			if (pch != NULL && (strcmp(pch, "GDT-TS-score=")==0)) {
				pch = strtok(NULL, " ");
				retVal = - atof(pch);
			}
		}
		return retVal;
	}

	/* Compute the distance between two PDB structures using popen */
	double slowDistance(TmObj & b) {
		char command[SZ_LINE];
		sprintf(command, "TMscore %s %s",
				fileName.c_str(),
				b.fileName.c_str());
		//system(command);
		FILE * fpipe;
		char line[256];

		if ( !(fpipe = (FILE*) popen(command, "r")) ) {
			perror("Problems with pipe!");
			exit(1);
		}

		double retVal = 0;
		while (fgets(line, sizeof line, fpipe)) {
			//string str = line;
			//retVal = getGdtTs(str);
			retVal = getGdtTs(line);

			if (retVal <= 0) {
				break;
			}
		}

		pclose(fpipe);
		return retVal;
	}

	/* Compute the distance between two PDB structures using direct function
	 * call to fortran code.
	 */
	double distance(TmObj & b) {
		char modelName[500];
		char nativeName[500];
		strcpy(modelName, fileName.c_str());
		strcpy(nativeName, b.fileName.c_str());
		return gdtts_(modelName, nativeName);
	}
};


#endif /* TMSCORE_H_ */
